require_relative './types.rb'
require_relative './utils.rb'

module Dhallish

	# Utility function to create dhallish function types:
	# Use [:name] to introduce a new unresolved type and write
	# :name to use it. Look at `fill_context` for examples.
	def make_fn_type(*args, rettype)
		if Types::not_a_type?(rettype); rettype = Types::Unresolved.new(rettype) end
		args.reverse.reduce(rettype) { |rettype, arg|
			argtype = arg
			name = nil
			if arg.is_a? Array
				name = arg[0]
				argtype = Types::Type.new(Types::Unresolved.new(name))
			end
			if Types::not_a_type?(argtype); argtype = Types::Unresolved.new(argtype) end
			Types::Function.new(argtype, rettype, name)
		}
	end
	module_function :make_fn_type

	def fill_context(globalctx, types)

		# Types:
		globalctx["Natural"] = Types::Natural
		globalctx["Integer"] = Types::Integer
		globalctx["Double"] = Types::Double
		globalctx["Bool"] = Types::Bool
		globalctx["Text"] = Types::Text
		globalctx["Type"] = Types::Type.new

		types["Natural"] = Types::Type.new(Types::Natural)
		types["Integer"] = Types::Type.new(Types::Integer)
		types["Double"] = Types::Type.new(Types::Double)
		types["Bool"] = Types::Type.new(Types::Bool)
		types["Text"] = Types::Type.new(Types::Text)
		types["Type"] = Types::Type.new(Types::Type.new)

		# Shows:
		globalctx["Natural/show"] = BuiltinFunction.new { |arg| arg.to_s }
		types["Natural/show"] = Types::Function.new(Types::Natural, Types::Text)

		globalctx["Integer/show"] = BuiltinFunction.new { |arg| arg.to_s }
		types["Integer/show"] = Types::Function.new(Types::Integer, Types::Text)

		globalctx["Double/show"] =  BuiltinFunction.new { |arg| arg.to_s }
		types["Double/show"] = Types::Function.new(Types::Double, Types::Text)

		globalctx["Text/show"] = BuiltinFunction.new { |arg| arg }
		types["Text/show"] = Types::Function.new(Types::Text, Types::Text)

		# Casts:
		natToInt = BuiltinFunction.new { |arg| arg }
		globalctx["Natural/toInteger"] = natToInt
		types["Natural/toInteger"] = Types::Function.new(Types::Natural, Types::Integer)

		natToDou = BuiltinFunction.new { |arg| arg.to_f }
		globalctx["Natural/toDouble"] = natToDou
		types["Natural/toDouble"] = Types::Function.new(Types::Natural, Types::Double)


		intToNat = BuiltinFunction.new { |arg| arg.abs  }
		globalctx["Integer/toNatural"] = intToNat
		types["Integer/toNatural"] = Types::Function.new(Types::Integer, Types::Natural)

		intToDou = BuiltinFunction.new { |arg| arg.to_f }
		globalctx["Integer/toDouble"] = intToDou
		types["Integer/toDouble"] = Types::Function.new(Types::Integer, Types::Double)


		douToNat = BuiltinFunction.new { |arg| arg.round.abs }
		globalctx["Double/toNatural"] = douToNat
		types["Double/toNatural"] = Types::Function.new(Types::Double, Types::Natural)

		douToInt = BuiltinFunction.new { |arg| arg.round }
		globalctx["Double/toInteger"] = douToInt
		types["Double/toInteger"] = Types::Function.new(Types::Double, Types::Integer)


		# Lists:
		globalctx["List"] = BuiltinFunction.new { |a| Types::List.new(a) }
		types["List"] = make_fn_type([:list_a], Types::Type.new(Types::List.new(Types::Unresolved.new(:list_a))))

		list_length_type = make_fn_type([:list_len_a], Types::List.new(Types::Unresolved.new(:list_len_a)), Types::Natural)
		globalctx["List/length"] = BuiltinFunction.new{ |a|
			BuiltinFunction.new { |list|
				list.length
			}
		}
		types["List/length"] = list_length_type

		list_fold_type = make_fn_type(
			[:list_fold_a],
			Types::List.new(Types::Unresolved.new(:list_fold_a)),
			[:list_fold_b],
			make_fn_type(:list_fold_a, :list_fold_b, :list_fold_b),
			:list_fold_b, :list_fold_b)
		globalctx["List/fold"] = BuiltinFunction.new { |a|
			BuiltinFunction.new { |list|
				BuiltinFunction.new { |b|
					BuiltinFunction.new { |f|
						BuiltinFunction.new { |x|
							acc = x
							list.reverse_each { |elm| acc = f.call(elm).call(acc) }
							acc
						}
					}
				}
			}
		}
		types["List/fold"] = list_fold_type

		list_head_type = make_fn_type(
			[:list_head_a],
			Types::List.new(Types::Unresolved.new(:list_head_a)),
			Types::Optional.new(Types::Unresolved.new(:list_head_a)))
		globalctx["List/head"] = BuiltinFunction.new { |a|
			BuiltinFunction.new { |list|
				list.first
			}
		}
		types["List/head"] = list_head_type

		list_last_type = make_fn_type(
			[:list_last_a],
			Types::List.new(Types::Unresolved.new(:list_last_a)),
			Types::Optional.new(Types::Unresolved.new(:list_last_a)))
		globalctx["List/last"] = BuiltinFunction.new { |a|
			BuiltinFunction.new { |list|
				list.last
			}
		}
		types["List/last"] = list_last_type

		list_tail_type = make_fn_type(
			[:list_tail_a],
			Types::List.new(Types::Unresolved.new(:list_tail_a)),
			Types::List.new(Types::Unresolved.new(:list_tail_a)))
		globalctx["List/tail"] = BuiltinFunction.new { |a|
			BuiltinFunction.new { |list|
				if list.empty?
					[]
				else
					list[1..list.length]
				end
			}
		}
		types["List/tail"] = list_tail_type

		types["List/reverse"] = make_fn_type(
			[:list_reverse_a],
			Types::List.new(Types::Unresolved.new(:list_reverse_a)),
			Types::List.new(Types::Unresolved.new(:list_reverse_a)))
		globalctx["List/reverse"] = BuiltinFunction.new { |a|
			BuiltinFunction.new { |list| list.reverse }
		}

		types["List/build"] = make_fn_type(
			[:list_build_a],
			make_fn_type(
				[:list_build_b],
				make_fn_type(:list_build_a, :list_build_b, :list_build_b), :list_build_b, :list_build_b),
			Types::List.new(Types::Unresolved.new(:list_build_a)))
		globalctx["List/build"] = BuiltinFunction.new { |a|
			BuiltinFunction.new { |f|
				cons = BuiltinFunction.new { |x|
					BuiltinFunction.new { |list| [x] + list }
				}
				f.call(Types::List.new(a)).call(cons).call([])
			}
		}

		types["List/indexed"] = make_fn_type(
			[:list_indexed_a],
			Types::List.new(Types::Unresolved.new(:list_indexed_a)),
			Types::List.new(Types::Record.new({
				"index" => Types::Natural,
				"value" => Types::Unresolved.new(:list_indexed_a)})))
		globalctx["List/indexed"] = BuiltinFunction.new { |a|
			BuiltinFunction.new { |list|
				list.map.with_index { |val, idx|
					{ "index" => idx, "value" => val }
				}
			}
		}

		# Optionals:
		globalctx["Optional"] = BuiltinFunction.new { |a|
			Types::Optional.new(a)
		}
		types["Optional"] = make_fn_type([:opt_a], Types::Type.new(Types::Optional.new(Types::Unresolved.new(:opt_a))))

		optional_fold_type = make_fn_type(
			[:opt_fold_a],
			Types::Optional.new(Types::Unresolved.new(:opt_fold_a)),
			[:opt_fold_b],
			make_fn_type(:opt_fold_a, :opt_fold_b),
			:opt_fold_b, :opt_fold_b)
		globalctx["Optional/fold"] = BuiltinFunction.new { |a|
			BuiltinFunction.new { |opt|
				BuiltinFunction.new { |b|
					BuiltinFunction.new { |f|
						BuiltinFunction.new { |x|
							if opt.nil?
								x
							else
								f.call(opt)
							end
						}
					}
				}
			}
		}
		types["Optional/fold"] = optional_fold_type

		types["Optional/build"] = make_fn_type(
			[:opt_build_a],
			make_fn_type([:opt_build_b],
				make_fn_type(:opt_build_a, :opt_build_b), :opt_build_b, :opt_build_b),
			Types::Optional.new(Types::Unresolved.new(:opt_build_a)))
		globalctx["Optional/build"] = BuiltinFunction.new { |a|
			BuiltinFunction.new { |f|
				just = BuiltinFunction.new { |x| x }
				f.call(Types::Optional.new(a)).call(just).call(nil)
			}
		}


		# Naturals:
		natural_fold_type = make_fn_type(
			Types::Natural,
			[:nat_fold_a],
			make_fn_type(:nat_fold_a, :nat_fold_a),
			:nat_fold_a, :nat_fold_a)
		globalctx["Natural/fold"] = BuiltinFunction.new { |n|
			BuiltinFunction.new { |a|
				BuiltinFunction.new { |succ|
					BuiltinFunction.new { |zero|
						res = zero
						n.times { res = succ.call(res) }
						res
					}
				}
			}
		}
		types["Natural/fold"] = natural_fold_type

		types["Natural/even"] = make_fn_type(Types::Natural, Types::Bool)
		globalctx["Natural/even"] = BuiltinFunction.new { |n| n % 2 == 0 }

		types["Natural/odd"] = make_fn_type(Types::Natural, Types::Bool)
		globalctx["Natural/odd"] = BuiltinFunction.new { |n| n % 2 == 1 }

		types["Natural/isZero"] = make_fn_type(Types::Natural, Types::Bool)
		globalctx["Natural/isZero"] = BuiltinFunction.new { |n| n == 0 }

		globalctx["Natural/build"] = BuiltinFunction.new { |f|
			zero = 0
			succ = BuiltinFunction.new { |n| n + 1 }
			f.call(Types::Natural).call(succ).call(zero)
		}
		types["Natural/build"] = make_fn_type(make_fn_type([:nat_build_a], make_fn_type(:nat_build_a, :nat_build_a), :nat_build_a, :nat_build_a), Types::Natural)

	end
	module_function :fill_context
end
